case_study = """path where the case study directory where custom classes for
organism, environment, experiment, observation have been specified in a mod.py
module and a simulation has been specified in a sim.py module which creates and
runs a simulation"""

scenario = """a string of the scenario specified for the case studies, which
contains folders with the config files for the scenario like events.csv,
config.json and sbi.json"""

package = """path relative to working directory ($PWD) where case studies are 
located"""